@@ -6,14 +6,14 @@ from .trainer import *
 from .protocol import *
 from .constants import *
 from .data import *
+from .pipeline import *
+from .zoo import *
 
 from .api import cv
-from .api import dl
 from .api import ml
 from .api import nlp
 from .api import multimodal
-from .api import interface as api
-from .api.zoo import *
+from .api import api
 from . import dist
 from . import scripts
 
