@@ -1,30 +1,27 @@
-import math
-
 import torch.nn as nn
 
+from torch import Tensor
 from typing import Any
 from typing import List
 from typing import Union
 from typing import Optional
 from cftool.misc import shallow_copy_dict
+from cftool.types import tensor_dict_type
 
-from .protocol import DecoderBase
-from .protocol import Decoder1DBase
-from ....types import tensor_dict_type
-from ....protocol import TrainerState
+from .protocol import DecoderMixin
+from .protocol import Decoder1DMixin
 from ....constants import INPUT_KEY
-from ....constants import PREDICTIONS_KEY
 from ....modules.blocks import get_conv_blocks
-from ....modules.blocks import Conv2d
 from ....modules.blocks import Lambda
 from ....modules.blocks import Linear
+from ....modules.blocks import Activation
 from ....modules.blocks import ResidualBlock
 from ....modules.blocks import ChannelPadding
 from ....modules.blocks import UpsampleConv2d
 
 
-@DecoderBase.register("vanilla")
-class VanillaDecoder(DecoderBase):
+@DecoderMixin.register("vanilla")
+class VanillaDecoder(nn.Module, DecoderMixin):
     def __init__(
         self,
         latent_channels: int,
@@ -45,16 +42,18 @@ class VanillaDecoder(DecoderBase):
         cond_channels: int = 16,
         num_classes: Optional[int] = None,
         latent_resolution: Optional[int] = None,
+        final_activation: Optional[str] = None,
     ):
-        super().__init__(
-            latent_channels,
-            out_channels,
+        super().__init__()
+        self.latent_channels = latent_channels
+        self._initialize(
+            out_channels=out_channels,
             img_size=img_size,
             num_upsample=num_upsample,
-            cond_channels=cond_channels,
             num_classes=num_classes,
             latent_resolution=latent_resolution,
         )
+        self._init_cond(cond_channels=cond_channels)
 
         blocks: List[nn.Module] = []
 
@@ -141,14 +140,12 @@ class VanillaDecoder(DecoderBase):
             in_nc = out_nc
 
         self.decoder = nn.ModuleList(blocks)
+        if final_activation is None:
+            self.final_activation = None
+        else:
+            self.final_activation = Activation.make(final_activation)
 
-    def forward(
-        self,
-        batch_idx: int,
-        batch: tensor_dict_type,
-        state: Optional[TrainerState] = None,
-        **kwargs: Any,
-    ) -> tensor_dict_type:
+    def forward(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
         net = self._inject_cond(batch)[INPUT_KEY]
         determinate = kwargs.pop("determinate", False)
         for block in self.decoder:
@@ -157,15 +154,18 @@ class VanillaDecoder(DecoderBase):
                 kw["determinate"] = determinate
             net = block(net, **kw)
         net = self.resize(net, determinate=determinate)
-        return {PREDICTIONS_KEY: net}
+        if self.final_activation is not None:
+            net = self.final_activation(net)
+        return net
 
 
-@Decoder1DBase.register("vanilla")
-class VanillaDecoder1D(Decoder1DBase):
+@Decoder1DMixin.register("vanilla")
+class VanillaDecoder1D(nn.Module, Decoder1DMixin):
     def __init__(
         self,
         latent_dim: int,
         out_channels: int,
+        latent_channels: int = 64,
         norm_type: Optional[str] = "instance",
         res_norm_type: Optional[str] = "instance",
         activation: Optional[str] = "leaky_relu_0.2",
@@ -182,12 +182,13 @@ class VanillaDecoder1D(Decoder1DBase):
         cond_channels: int = 16,
         num_classes: Optional[int] = None,
         latent_resolution: Optional[int] = None,
-        latent_padding_channels: Optional[int] = 16,
-        latent_expand_ratio: int = 1,
+        latent_padding_channels: Optional[int] = None,
+        final_activation: Optional[str] = None,
     ):
-        super().__init__(
-            latent_dim,
-            out_channels,
+        super().__init__()
+        self.latent_dim = latent_dim
+        self._initialize(
+            out_channels=out_channels,
             img_size=img_size,
             num_upsample=num_upsample,
             num_classes=num_classes,
@@ -196,26 +197,24 @@ class VanillaDecoder1D(Decoder1DBase):
         if latent_resolution is None:
             latent_resolution = int(round(img_size / 2**self.num_upsample))
         self.latent_resolution = latent_resolution
-        in_dim = latent_dim
-        latent_dim *= latent_expand_ratio
+        assert isinstance(self.latent_resolution, int)
         latent_area = self.latent_resolution**2
-        latent_channels = math.ceil(latent_dim / latent_area)
         shape = -1, latent_channels, latent_resolution, latent_resolution
         blocks: List[nn.Module] = [
-            Linear(in_dim, latent_channels * latent_area),
+            Linear(latent_dim, latent_channels * latent_area),
+            Activation.make(activation),
             Lambda(lambda tensor: tensor.view(*shape), f"reshape -> {shape}"),
-            Conv2d(latent_channels, latent_dim, kernel_size=1),
         ]
         if latent_padding_channels is not None:
             latent_padding = ChannelPadding(
-                latent_dim,
+                latent_channels,
                 latent_padding_channels,
                 latent_resolution,
             )
             blocks.append(latent_padding)
         self.from_latent = nn.Sequential(*blocks)
         self.decoder = VanillaDecoder(
-            latent_dim,
+            latent_channels,
             out_channels,
             norm_type,
             res_norm_type,
@@ -232,20 +231,15 @@ class VanillaDecoder1D(Decoder1DBase):
             cond_channels=cond_channels,
             num_classes=num_classes,
             latent_resolution=latent_resolution,
+            final_activation=final_activation,
         )
 
-    def forward(
-        self,
-        batch_idx: int,
-        batch: tensor_dict_type,
-        state: Optional[TrainerState] = None,
-        **kwargs: Any,
-    ) -> tensor_dict_type:
+    def forward(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
         batch = shallow_copy_dict(batch)
         net = batch[INPUT_KEY]
         net = self.from_latent(net)
         batch[INPUT_KEY] = net
-        return self.decoder(batch_idx, batch, state, **kwargs)
+        return self.decoder(batch, **kwargs)
 
 
 __all__ = [
