@@ -10,7 +10,6 @@ import torch.distributed as dist
 from typing import Any
 from typing import Dict
 from typing import List
-from typing import Type
 from typing import Tuple
 from typing import Union
 from typing import Callable
@@ -18,39 +17,44 @@ from typing import Optional
 from typing import NamedTuple
 from tqdm.autonotebook import tqdm
 from torch.optim import Optimizer
+from cftool.misc import print_info
+from cftool.misc import print_error
 from cftool.misc import update_dict
+from cftool.misc import print_warning
 from cftool.misc import shallow_copy_dict
 from cftool.misc import sort_dict_by_value
 from cftool.misc import context_error_handler
-from cftool.misc import WithRegister
+from cftool.array import to_device
+from cftool.types import tensor_dict_type
 from torch.optim.lr_scheduler import _LRScheduler
 from torch.nn.parallel import DistributedDataParallel as DDP
 from torch.distributed.optim import ZeroRedundancyOptimizer as ZeRO
 from torch.utils.data.distributed import DistributedSampler
 
 from .data import DLLoader
-from .data import DLDataModule
-from .types import tensor_dict_type
+from .protocol import ITrainer
+from .protocol import DeviceInfo
+from .protocol import IDataModule
 from .protocol import StepOutputs
-from .protocol import LossProtocol
+from .protocol import ILoss
+from .protocol import TqdmSettings
 from .protocol import TrainerState
-from .protocol import ModelProtocol
+from .protocol import IDLModel
+from .protocol import OptimizerPack
 from .protocol import MetricsOutputs
 from .protocol import MonitorResults
 from .protocol import TrainerMonitor
-from .protocol import MetricProtocol
-from .protocol import InferenceProtocol
-from .protocol import DataLoaderProtocol
+from .protocol import _IMetric
+from .protocol import _MultipleMetrics
+from .protocol import TrainerCallback
+from .protocol import IInference
+from .protocol import IDataLoader
 from .protocol import ModelWithCustomSteps
 from .constants import LOSS_KEY
 from .constants import PT_PREFIX
-from .constants import INFO_PREFIX
 from .constants import SCORES_FILE
-from .constants import ERROR_PREFIX
-from .constants import WARNING_PREFIX
 from .constants import CHECKPOINTS_FOLDER
 from .misc.toolkit import summary
-from .misc.toolkit import to_device
 from .misc.toolkit import get_ddp_info
 from .misc.toolkit import has_batch_norms
 from .misc.toolkit import scheduler_requires_metric
@@ -58,21 +62,9 @@ from .misc.toolkit import eval_context
 from .modules.optimizers import optimizer_dict
 from .modules.schedulers import scheduler_dict
 from .modules.schedulers import WarmupScheduler
-from .misc.internal_.metrics import MultipleMetrics
 from .misc.internal_.monitors import ConservativeMonitor
 
 
-callback_dict: Dict[str, Type["TrainerCallback"]] = {}
-
-
-class OptimizerPack(NamedTuple):
-    scope: str
-    optimizer_name: str
-    scheduler_name: Optional[str] = None
-    optimizer_config: Optional[Dict[str, Any]] = None
-    scheduler_config: Optional[Dict[str, Any]] = None
-
-
 class amp_context(context_error_handler):
     def __init__(self, enabled: bool):
         if not enabled:
@@ -141,84 +133,6 @@ class DefaultOptimizerSettings(NamedTuple):
         )
 
 
-class DeviceInfo(NamedTuple):
-    cuda: Optional[str]
-    rank: Optional[int]
-
-    @property
-    def device(self) -> torch.device:
-        if self.rank is not None:
-            return torch.device(f"cuda:{self.rank}")
-        return torch.device("cpu" if self.cuda is None else f"cuda:{self.cuda}")
-
-
-class TrainerCallback(WithRegister["TrainerCallback"]):
-    d = callback_dict
-    is_rank_0: bool = True
-
-    def __init__(self, *args: Any, **kwargs: Any):
-        pass
-
-    def initialize(self) -> None:
-        pass
-
-    def mutate_train_forward_kwargs(
-        self,
-        kwargs: Dict[str, Any],
-        trainer: "Trainer",
-    ) -> None:
-        pass
-
-    def mutate_train_loss_kwargs(
-        self,
-        kwargs: Dict[str, Any],
-        trainer: "Trainer",
-    ) -> None:
-        pass
-
-    def before_loop(self, trainer: "Trainer") -> None:
-        pass
-
-    def log_lr(self, key: str, lr: float, state: TrainerState) -> None:
-        pass
-
-    def log_metrics(self, metric_outputs: MetricsOutputs, state: TrainerState) -> None:
-        pass
-
-    def log_metrics_msg(
-        self,
-        metric_outputs: MetricsOutputs,
-        metrics_log_path: str,
-        state: TrainerState,
-    ) -> None:
-        pass
-
-    def log_artifacts(self, trainer: "Trainer") -> None:
-        pass
-
-    def after_step(self, step_outputs: StepOutputs, state: TrainerState) -> None:
-        pass
-
-    def after_monitor(
-        self,
-        monitor_results: MonitorResults,
-        state: TrainerState,
-    ) -> None:
-        pass
-
-    def finalize(self, trainer: "Trainer") -> None:
-        pass
-
-
-class TqdmSettings(NamedTuple):
-    use_tqdm: bool = False
-    use_step_tqdm: bool = False
-    use_tqdm_in_validation: bool = False
-    in_distributed: bool = False
-    position: int = 0
-    desc: str = "epoch"
-
-
 def get_sorted_checkpoints(checkpoint_folder: str) -> List[str]:
     # better checkpoints will be placed earlier,
     #  which means `checkpoints[0]` is the best checkpoint
@@ -241,19 +155,7 @@ def _setup_ddp(*, backend: str = "nccl", dist_url: str = "env://") -> None:
     )
 
 
-class Trainer:
-    loss: LossProtocol
-    model: ModelProtocol
-    metrics: Optional[MetricProtocol]
-    monitors: List[TrainerMonitor]
-    callbacks: List[TrainerCallback]
-    state: TrainerState
-    device_info: DeviceInfo
-    train_loader: DataLoaderProtocol
-    train_loader_copy: DataLoaderProtocol
-    valid_loader: Optional[DataLoaderProtocol]
-    inference: InferenceProtocol
-
+class Trainer(ITrainer):
     def __init__(
         self,
         state_config: Optional[Dict[str, Any]] = None,
@@ -265,7 +167,7 @@ class Trainer:
         valid_portion: float = 1.0,
         amp: bool = False,
         clip_norm: float = 0.0,
-        metrics: Optional[MetricProtocol] = None,
+        metrics: Optional[_IMetric] = None,
         use_losses_as_metrics: Optional[bool] = None,
         loss_metrics_weights: Optional[Dict[str, float]] = None,
         recompute_train_losses_in_eval: bool = True,
@@ -276,6 +178,7 @@ class Trainer:
         scheduler_name: Optional[str] = None,
         optimizer_config: Optional[Dict[str, Any]] = None,
         scheduler_config: Optional[Dict[str, Any]] = None,
+        update_scheduler_per_epoch: bool = False,
         optimizer_packs: Optional[Union[OptimizerPack, List[OptimizerPack]]] = None,
         use_zero: bool = False,
         data_info_name: str = "data_info",
@@ -318,6 +221,8 @@ class Trainer:
             settings["optimizer_config"] = optimizer_config
         if scheduler_config is not None:
             settings["scheduler_config"] = scheduler_config
+        self.update_scheduler_per_epoch = update_scheduler_per_epoch
+        self._current_scheduler_epoch = -1
         self.default_opt_settings = DefaultOptimizerSettings(**settings)
         if optimizer_packs is None:
             self.optimizer_packs = None
@@ -347,7 +252,7 @@ class Trainer:
                 msg = "`metrics` should be provided when not `use_losses_as_metrics`"
                 raise ValueError(msg)
         else:
-            if not isinstance(metrics, MultipleMetrics):
+            if not isinstance(metrics, _MultipleMetrics):
                 metrics.trainer = self
             else:
                 for metric in metrics.metrics:
@@ -378,8 +283,8 @@ class Trainer:
         # initialize artifact structure
         self.checkpoint_folder = None
         self.data_info_name = data_info_name
+        self.workplace = workplace
         if self.is_rank_0:
-            self.workplace = workplace
             os.makedirs(self.workplace, exist_ok=True)
             self.metrics_log_path = os.path.join(self.workplace, metrics_log_file)
             with open(self.metrics_log_path, "w"):
@@ -392,6 +297,36 @@ class Trainer:
         self.final_results: Optional[MetricsOutputs] = None
         self.checkpoint_scores: Dict[str, float] = {}
 
+    @property
+    def config(self) -> Dict[str, Any]:
+        ddp_info = get_ddp_info()
+        ddp_d = None if ddp_info is None else ddp_info._asdict()
+        return {
+            "state_config": self.state.config,
+            "valid_portion": self.valid_portion,
+            "amp": self.use_amp,
+            "clip_norm": self.clip_norm,
+            "metrics": (
+                None
+                if self.metrics is None
+                else self.metrics.__identifier__
+                if not isinstance(self.metrics, _MultipleMetrics)
+                else [metric.__identifier__ for metric in self.metrics.metrics]
+            ),
+            "loss_metrics_weights": self.loss_metrics_weights,
+            "monitors": [monitor.__identifier__ for monitor in self.monitors],
+            "callbacks": [callback.__identifier__ for callback in self.callbacks],
+            "optimizer_packs": (
+                None
+                if self.optimizer_packs is None
+                else [pack._asdict() for pack in self.optimizer_packs]
+            ),
+            "ddp_info": ddp_d,
+            "finetune_config": self.finetune_config,
+            "tqdm_settings": self.tqdm_settings._asdict(),
+            "device_info": self.device_info._asdict(),
+        }
+
     @property
     def device(self) -> torch.device:
         return self.device_info.device
@@ -405,7 +340,7 @@ class Trainer:
         return self.tqdm_settings.use_tqdm_in_validation or self.state.is_terminate
 
     @property
-    def validation_loader(self) -> DataLoaderProtocol:
+    def validation_loader(self) -> IDataLoader:
         return self.valid_loader or self.train_loader_copy
 
     @property
@@ -462,14 +397,14 @@ class Trainer:
         pretrained_ckpt = self.finetune_config.get("pretrained_ckpt")
         if pretrained_ckpt is None:
             raise ValueError("`rank` should be provided when `finetune` is triggered")
-        print(f"{INFO_PREFIX}loading pretrained checkpoint from '{pretrained_ckpt}'...")
+        print_info(f"loading pretrained checkpoint from '{pretrained_ckpt}'...")
         d = torch.load(pretrained_ckpt, map_location=self.device)
         self.model.load_state_dict(d)
         freeze = self.finetune_config.get("freeze", "")
         freeze_except = self.finetune_config.get("freeze_except", "")
         if not freeze and not freeze_except:
             return None
-        msg_fmt = f"{INFO_PREFIX}{'{}'} parameters will be {'{}'} under '{'{}'}'"
+        msg_fmt = f"-> {'{}'} parameters will be {'{}'} under '{'{}'}'"
         param_names = []
         if freeze:
             num_frozen = 0
@@ -546,8 +481,8 @@ class Trainer:
         if pack.scope == "all":
             if self.model_has_custom_steps and self.model.custom_params_groups:
                 if self.use_zero:
-                    print(
-                        f"{WARNING_PREFIX}currently PyTorch does not support "
+                    print_warning(
+                        "currently PyTorch does not support "
                         "using ZeRO with parameter groups, so ZeRO will be disabled"
                     )
                     self.use_zero = False
@@ -630,15 +565,17 @@ class Trainer:
         loss = loss_dict[LOSS_KEY]
         self.grad_scaler.scale(loss).backward()
         # clip norm
-        if self.clip_norm > 0.0:
-            self.clip_norm_step()
+        self.clip_norm_step()
         # optimize
         self.optimizer_step()
         self.scheduler_step()
 
     def weighted_loss_score(self, loss_items: Dict[str, float]) -> float:
         if not self.loss_metrics_weights:
-            return -loss_items[LOSS_KEY]
+            loss = loss_items.get(LOSS_KEY)
+            if loss is not None:
+                return -loss
+            return -sum(loss_items.values()) / len(loss_items)
         score = 0.0
         for k, w in self.loss_metrics_weights.items():
             v = loss_items.get(k)
@@ -648,12 +585,13 @@ class Trainer:
         return score
 
     def clip_norm_step(self) -> None:
-        for opt in self.optimizers.values():
-            self.grad_scaler.unscale_(opt)
-        self._gradient_norm = nn.utils.clip_grad_norm_(
-            self.model_for_training.parameters(),
-            max_norm=self.clip_norm,
-        )
+        if self.clip_norm > 0.0:
+            for opt in self.optimizers.values():
+                self.grad_scaler.unscale_(opt)
+            self._gradient_norm = nn.utils.clip_grad_norm_(
+                self.model_for_training.parameters(),
+                max_norm=self.clip_norm,
+            )
 
     def optimizer_step(self) -> None:
         for opt in self.optimizers.values():
@@ -663,11 +601,14 @@ class Trainer:
             param.grad = None
 
     def scheduler_step(self) -> None:
+        if self.update_scheduler_per_epoch:
+            if self.state.epoch == self._current_scheduler_epoch:
+                return
         lr_metric_logged = False
         for key, scheduler in self.schedulers.items():
             if scheduler is not None:
                 should_log_lr, kwargs = self._get_scheduler_settings(key, scheduler)
-                if should_log_lr:
+                if should_log_lr or self.update_scheduler_per_epoch:
                     lr_metric_logged = True
                     for callback in self.callbacks:
                         callback.log_lr(
@@ -678,6 +619,8 @@ class Trainer:
                 scheduler.step(**shallow_copy_dict(kwargs))
         if lr_metric_logged:
             self.lr_metrics_updated = False
+        if self.update_scheduler_per_epoch:
+            self._current_scheduler_epoch = self.state.epoch
 
     def _get_scheduler_settings(
         self,
@@ -736,7 +679,9 @@ class Trainer:
             if self.state.should_start_snapshot:
                 score = self.intermediate.final_score
                 if any(monitor.snapshot(score) for monitor in self.monitors):
-                    save_checkpoint = True
+                    if self.state.can_snapshot:
+                        self.state.update_snapshot_epoch()
+                        save_checkpoint = True
                 if any(monitor.check_terminate(score) for monitor in self.monitors):
                     terminate = True
         return MonitorResults(terminate, save_checkpoint, self.intermediate)
@@ -769,42 +714,12 @@ class Trainer:
 
     # api
 
-    @property
-    def config(self) -> Dict[str, Any]:
-        ddp_info = get_ddp_info()
-        ddp_d = None if ddp_info is None else ddp_info._asdict()
-        return {
-            "state_config": self.state.config,
-            "valid_portion": self.valid_portion,
-            "amp": self.use_amp,
-            "clip_norm": self.clip_norm,
-            "metrics": (
-                None
-                if self.metrics is None
-                else self.metrics.__identifier__
-                if not isinstance(self.metrics, MultipleMetrics)
-                else [metric.__identifier__ for metric in self.metrics.metrics]
-            ),
-            "loss_metrics_weights": self.loss_metrics_weights,
-            "monitors": [monitor.__identifier__ for monitor in self.monitors],
-            "callbacks": [callback.__identifier__ for callback in self.callbacks],
-            "optimizer_packs": (
-                None
-                if self.optimizer_packs is None
-                else [pack._asdict() for pack in self.optimizer_packs]
-            ),
-            "ddp_info": ddp_d,
-            "finetune_config": self.finetune_config,
-            "tqdm_settings": self.tqdm_settings._asdict(),
-            "device_info": self.device_info._asdict(),
-        }
-
     def fit(
         self,
-        data: DLDataModule,
-        loss: LossProtocol,
-        model: ModelProtocol,
-        inference: InferenceProtocol,
+        data: IDataModule,
+        loss: ILoss,
+        model: IDLModel,
+        inference: IInference,
         *,
         config_export_file: Optional[str] = None,
         show_summary: Optional[bool] = None,
@@ -864,7 +779,7 @@ class Trainer:
         # train
         has_ckpt = terminate = False
         if self.is_rank_0 and self.epoch_tqdm is None:
-            print(f"{INFO_PREFIX}entered training loop")
+            print_info("entered training loop")
         if self.is_rank_0 and config_export_file is not None:
             config_export_path = os.path.join(self.workplace, config_export_file)
             with open(config_export_path, "w") as f:
@@ -906,7 +821,7 @@ class Trainer:
             except KeyboardInterrupt:
                 if dist.is_initialized():
                     raise
-                print(f"{ERROR_PREFIX}keyboard interrupted")
+                print_error("keyboard interrupted")
                 terminate = True
             if terminate:
                 break
@@ -920,7 +835,7 @@ class Trainer:
         # restore
         if self.is_rank_0 and self.has_checkpoint_folder:
             if not self.tqdm_settings.in_distributed:
-                print(f"{INFO_PREFIX}rolling back to the best checkpoint")
+                print_info("rolling back to the best checkpoint")
             has_ckpt = self.restore_checkpoint()
         # finalize
         self.state.set_terminate()
@@ -937,7 +852,7 @@ class Trainer:
         self,
         *,
         portion: float = 1.0,
-        loader: Optional[DataLoaderProtocol] = None,
+        loader: Optional[IDataLoader] = None,
     ) -> MetricsOutputs:
         if loader is None:
             loader = self.validation_loader
@@ -989,11 +904,13 @@ class Trainer:
         state = getattr(self, "state", None)
         pt_file = f"{PT_PREFIX}{-1 if state is None else state.step}.pt"
         if state is None:
-            print(
-                f"{WARNING_PREFIX}`state` is not initialized, "
-                "latest model will be saved and no scores will be recorded"
+            print_warning(
+                "`state` is not initialized, "
+                "latest model will be saved and the recorded score will always be 0"
             )
             torch.save(self.model.state_dict(), os.path.join(folder, pt_file))
+            with open(os.path.join(folder, SCORES_FILE), "w") as f:
+                json.dump({pt_file: 0.0}, f)
             return
         # leave top_k snapshots only
         if state.max_snapshot_file > 0:
@@ -1025,7 +942,7 @@ class Trainer:
         checkpoints = get_sorted_checkpoints(folder)
         if not checkpoints:
             if not self.tqdm_settings.in_distributed:
-                print(f"{WARNING_PREFIX}no model file found in {folder}")
+                print_warning(f"no model file found in {folder}")
             return False
         success = False
         for checkpoint in checkpoints:
@@ -1033,7 +950,7 @@ class Trainer:
             if not os.path.isfile(model_file):
                 continue
             if not self.tqdm_settings.in_distributed:
-                print(f"{INFO_PREFIX}restoring from {model_file}")
+                print_info(f"restoring from {model_file}")
             states = torch.load(model_file, map_location=self.device)
             if state_dict_callback is not None:
                 state_dict_callback(states)
@@ -1046,10 +963,4 @@ class Trainer:
 __all__ = [
     "get_sorted_checkpoints",
     "Trainer",
-    "DeviceInfo",
-    "TqdmSettings",
-    "callback_dict",
-    "TrainerCallback",
-    "StepOutputs",
-    "OptimizerPack",
 ]
